/***************************************************************************
* Copyright (C)1999 The XAOL Project, All Rights Reserved.                 *
* WWW: http://www.xaol.org/                                                *
***************************************************************************/
// DataStream.c

#include "DataStream.h"

#include <stdlib.h>
#include <string.h>

DataStream* DSNew (char* buffer, int length, int byteSwap) {	
	DataStream* ds = NULL;
	
	if (length <= 0)
		return (NULL);
		
	ds = (DataStream*) MAlloc(sizeof(DataStream));
	XAOL_AssertIfNULL_(ds);
	if (ds != NULL) {
		ds->buffer = (char*) MAlloc(length);
		XAOL_AssertIfNULL_(ds->buffer);
		if (ds->buffer != NULL) {
			if (buffer != NULL)
				memcpy(ds->buffer, buffer, length);
			else
				memset(ds->buffer, 0, length);
			ds->p = ds->buffer;
			ds->position = 0;
			ds->end = length;
			ds->byteSwap = byteSwap;
			ds->checkValue = 131;
			}
		else {			
			MFree(ds);
			ds = NULL;
			}
		}
	
	return (ds);
	}

int DSIsValid (DataStream* ds) {
	return (ds != NULL && ds->checkValue == 131);
	}
	
void DSDispose (DataStream* ds) {
	XAOL_AssertIfNULL_(ds);
	if (DSIsValid(ds)) {
		if (ds->buffer != NULL)
			MFree(ds->buffer);
		MFree(ds);
		}
	}
	
XAOLError DSGetBuffer (DataStream* ds, char** buffer) {
	XAOL_AssertIfNULL_(ds);
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
	
	if (*buffer == NULL) {
		*buffer = (char*) MAlloc(ds->position);
		if (! *buffer)
			return (Mem_Error);
		}
	else {
		*buffer = (char*) MRealloc((void*) *buffer, ds->position);
		if (! *buffer)
			return (Mem_Error);
		}
	
	memcpy(*buffer, ds->buffer, DSGetUsed(ds));
	
	return (No_Error);
	}

XAOLError DSGetRaw(DataStream* ds, unsigned char** buffer, int length) {
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + length) > ds->end || length <= 0)
		return (DE_InvalidPos);

	*buffer = (unsigned char*) MAlloc(length);
	if (*buffer != NULL) {
		memcpy(*buffer, ds->p, length);
		ds->p  += length;
		ds->position += length;
		}

	return (No_Error);
	}

XAOLError DSPutRaw(DataStream* ds, unsigned char* data, int length) {
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + length) > ds->end)
		return (DE_InvalidPos);

	memcpy(ds->p, data, length);
	ds->p  += length;
	ds->position += length;

	return (No_Error);
	}
	
XAOLError DSGetByte (DataStream* ds, unsigned char* result) {
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + sizeof(unsigned char)) > ds->end)
		return (DE_InvalidPos);
	
	memcpy((unsigned char *) result, ds->p, sizeof(unsigned char));
	ds->p += sizeof(unsigned char);
	ds->position += sizeof(unsigned char);
	
	return (No_Error);
	}

XAOLError DSPutByte (DataStream* ds, unsigned char value) {
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
	
	if ((ds->position + sizeof(unsigned char)) > ds->end)
		return (DE_InvalidPos);

	memcpy(ds->p, (unsigned char *)&value, sizeof(unsigned char));
	ds->p  += sizeof(unsigned char);
	ds->position += sizeof(unsigned char);
	
	return (No_Error);
	}

XAOLError DSGetString (DataStream* ds, char** result) {
	int length = strlen(ds->p) + 1;
	
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + length) > ds->end || length <= 0)
		return (DE_InvalidPos);

	*result = (char*) MAlloc(length);
	if (*result != NULL) {
		memcpy(*result, ds->p, length);
		ds->p  += length;
		ds->position += length;
		}

	return (No_Error);
	}

XAOLError DSPutString (DataStream* ds, char* value) {
	int len = strlen(value) + 1;
	
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
	
	if ((ds->position + strlen(value) + 1) > ds->end || len <= 0)
		return (DE_InvalidPos);

	memcpy(ds->p, value, len);
	ds->p  += sizeof(len);
	ds->position += sizeof(len);

	return (No_Error);
	}

XAOLError DSGetUInt8 (DataStream* ds, UInt8* result) {
	XAOL_AssertIfNULL_(ds);
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + sizeof(UInt8)) > ds->end)
		return (DE_InvalidPos);
	
	memcpy((unsigned char *) result, ds->p, sizeof(UInt8));
	ds->p += sizeof(UInt8);
	ds->position += sizeof(UInt8);
	
	if (ds->byteSwap)
		*result = ntohs(*result);
		
	return (No_Error);
	}

XAOLError DSPutUInt8 (DataStream* ds, UInt8 value) {
	UInt8 ival;
	
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if (ds->byteSwap)
		ival = htons(value);
	else
		ival = value;
	
	if ((ds->position + sizeof(UInt8)) > ds->end)
		return (DE_InvalidPos);

	memcpy(ds->p, (unsigned char *)&ival, sizeof(UInt8));
	ds->p  += sizeof(UInt8);
	ds->position += sizeof(UInt8);

	return (No_Error);
	}


XAOLError DSGetUInt16 (DataStream* ds, UInt16* result) {
	XAOL_AssertIfNULL_(ds);
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + sizeof(UInt16)) > ds->end)
		return (DE_InvalidPos);
	
	memcpy((unsigned char *) result, ds->p, sizeof(UInt16));
	ds->p += sizeof(UInt16);
	ds->position += sizeof(UInt16);
	
	if (ds->byteSwap)
		*result = ntohs(*result);
		
	return (No_Error);
	}

XAOLError DSPutUInt16 (DataStream* ds, UInt16 value) {
	UInt16 ival;
	
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if (ds->byteSwap)
		ival = htons(value);
	else
		ival = value;
	
	if ((ds->position + sizeof(UInt16)) > ds->end)
		return (DE_InvalidPos);

	memcpy(ds->p, (unsigned char *)&ival, sizeof(UInt16));
	ds->p  += sizeof(UInt16);
	ds->position += sizeof(UInt16);

	return (No_Error);
	}

XAOLError DSGetUInt32 (DataStream* ds, UInt32* result) {
	XAOL_AssertIfNULL_(ds);
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + sizeof(UInt32)) > ds->end)
		return (DE_InvalidPos);
	
	memcpy((unsigned char *) result, ds->p, sizeof(UInt32));
	ds->p += sizeof(UInt32);
	ds->position += sizeof(UInt32);
	
	if (ds->byteSwap)
		*result = ntohl(*result);
		
	return (No_Error);
	}

XAOLError DSPutUInt32 (DataStream* ds, UInt32 value) {
	UInt32 ival;
	
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if (ds->byteSwap)
		ival = htonl(value);
	else
		ival = value;
	
	if ((ds->position + sizeof(UInt32)) > ds->end)
		return (DE_InvalidPos);

	memcpy(ds->p, (unsigned char *)&ival, sizeof(UInt32));
	ds->p  += sizeof(UInt32);
	ds->position += sizeof(UInt32);

	return (No_Error);
	}

XAOLError DSGetSInt8 (DataStream* ds, SInt8* result) {
	XAOL_AssertIfNULL_(ds);
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + sizeof(SInt8)) > ds->end)
		return (DE_InvalidPos);
	
	memcpy((unsigned char *) result, ds->p, sizeof(SInt8));
	ds->p += sizeof(SInt8);
	ds->position += sizeof(SInt8);
	
	if (ds->byteSwap)
		*result = ntohs(*result);
		
	return (No_Error);
	}

XAOLError DSPutSInt8 (DataStream* ds, SInt8 value) {
	SInt8 ival;
	
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if (ds->byteSwap)
		ival = htons(value);
	else
		ival = value;
	
	if ((ds->position + sizeof(SInt8)) > ds->end)
		return (DE_InvalidPos);

	memcpy(ds->p, (unsigned char *)&ival, sizeof(SInt8));
	ds->p  += sizeof(SInt8);
	ds->position += sizeof(SInt8);

	return (No_Error);
	}
	
XAOLError DSGetSInt16 (DataStream* ds, SInt16* result) {
	XAOL_AssertIfNULL_(ds);
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + sizeof(SInt16)) > ds->end)
		return (DE_InvalidPos);
	
	memcpy((unsigned char *) result, ds->p, sizeof(SInt16));
	ds->p += sizeof(SInt16);
	ds->position += sizeof(SInt16);
	
	if (ds->byteSwap)
		*result = ntohs(*result);
		
	return (No_Error);
	}

XAOLError DSPutSInt16 (DataStream* ds, SInt16 value) {
	SInt16 ival;
	
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if (ds->byteSwap)
		ival = htons(value);
	else
		ival = value;
	
	if ((ds->position + sizeof(SInt16)) > ds->end)
		return (DE_InvalidPos);

	memcpy(ds->p, (unsigned char *)&ival, sizeof(SInt16));
	ds->p  += sizeof(SInt16);
	ds->position += sizeof(SInt16);

	return (No_Error);
	}

XAOLError DSGetSInt32 (DataStream* ds, SInt32* result) {
	XAOL_AssertIfNULL_(ds);
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + sizeof(SInt32)) > ds->end)
		return (DE_InvalidPos);
	
	memcpy((unsigned char *) result, ds->p, sizeof(SInt32));
	ds->p += sizeof(SInt32);
	ds->position += sizeof(SInt32);
	
	if (ds->byteSwap)
		*result = ntohl(*result);
		
	return (No_Error);
	}

XAOLError DSPutSInt32 (DataStream* ds, SInt32 value) {
	SInt32 ival;
	
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if (ds->byteSwap)
		ival = htonl(value);
	else
		ival = value;
	
	if ((ds->position + sizeof(SInt32)) > ds->end)
		return (DE_InvalidPos);

	memcpy(ds->p, (unsigned char *)&ival, sizeof(SInt32));
	ds->p  += sizeof(SInt32);
	ds->position += sizeof(SInt32);

	return (No_Error);
	}
	
XAOLError DSPutDS (DataStream* ds, DataStream* append) {
	int len;
	
	if (DSIsValid(ds) == false || DSIsValid(append) == false)
		return (DE_InvalidDS);
	
	len = DSGetLength(append);
	
	if ((ds->position + len) > ds->end)
		return (DE_InvalidPos);
	
	memcpy(ds->p, (unsigned char *)append->buffer, len);
	ds->p  += len;
	ds->position += len;

	return (No_Error);
	}
	
int DSGetLength (DataStream* ds) {
	if (DSIsValid(ds))
		return (ds->end);
	else
		return (-1);
	}

int DSGetAvailable (DataStream* ds) {
	if (DSIsValid(ds))
		return (ds->end - ds->position);
	else
		return (-1);
	}

int DSGetUsed (DataStream* ds) {
	if (DSIsValid(ds))
		return (ds->position);
	else
		return (-1);
	}

XAOLError DSAdvance (DataStream* ds, int byteCount) {
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position + byteCount) > ds->end)
		return (DE_InvalidPos);
		
	ds->position += byteCount;
	ds->p += byteCount;
	
	return (No_Error);
	}
	
XAOLError DSRewind (DataStream* ds, int byteCount) {
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	if ((ds->position - byteCount) < 0)
		return (DE_InvalidPos);
		
	ds->position -= byteCount;
	ds->p -= byteCount;
	
	return (No_Error);
	}

XAOLError DSReset (DataStream* ds) {
	if (! DSIsValid(ds))
		return (DE_InvalidDS);
		
	ds->p = ds->buffer;
	ds->position = 0;
	
	return (No_Error);
	}