/***************************************************************************
* Copyright (C)1999 The XAOL Project, All Rights Reserved.                 *
* WWW: http://www.xaol.org/                                                *
***************************************************************************/
// P3.c

#include "P3.h"
#include "DataStream.h"
#include "CRC.h"
#include "ByteManip.h"

#include <stdlib.h>
#include <string.h>

// Create a new P3 Instance
P3* P3New (char type) {
	P3* p3 = NULL;
	
	p3 = (P3*) MAlloc(sizeof(P3));
	if (p3 != NULL) {		
		p3->checkValue 	= 129;
		p3->type   		= type;

		p3->seq    		= p3->ack = 0;
		p3->length		= 0;
		p3->data   		= NULL;
		}
	
	return (p3);
	}

// Perform basic sanity checks
int P3IsValid (P3* p3) {		
	return (p3 != NULL && p3->checkValue == 129);
	}
	
XAOLError P3Copy (P3* source, P3* dest, int copyData) {
	XAOL_AssertIfNULL_(source);
		
	if (! P3IsValid(source))
		return (PE_InvalidP3);
	
	if (dest != NULL || P3IsValid(dest) == false)
		dest = (P3*) MAlloc(sizeof(P3));
	
	XAOL_AssertIfNULL_(dest);
	
	if (P3IsValid(dest)) {
		dest->crc = source->crc;
  		dest->length = source->length;
		dest->seq  = source->seq;
		dest->ack  = source->ack;
		dest->type = source->type;
		
		if (copyData && source->length > 0) {
			if (dest->data != NULL)
				dest->data = (char*) MRealloc((void*) dest->data, source->length);
			else
				dest->data = (char*) MAlloc (source->length);				
			memcpy(dest->data, source->data, source->length);
			}			
		else
			dest->data = NULL;
		}
	else
		return (Mem_Error);
	
	return (No_Error);
	}

XAOLError P3Append (P3* p3, char* data, int len) {
	XAOL_AssertIfNULL_(p3);
	
	if (! P3IsValid(p3)) {
		XAOL_DEBUG_Print_("[XAOL] P3Append : Invalid P3");
		return (PE_InvalidP3);
		}
		
	if (data == NULL || len <= 0){
		XAOL_DEBUG_Print_("[XAOL] P3Append : No Data to append");
		return (PE_NoDataToAppend);
		}
		
	p3->data = (char*) MRealloc(p3->data, p3->length + len);
	memcpy(p3->data + p3->length, data, len);
	p3->length += len;

	return (No_Error);
	}

XAOLError P3AppendChar (P3* p3, char c) {
	char buffer[1];	
	buffer[0] = c;
	
	return (P3Append(p3, buffer, 1));
	}

void P3Dispose (P3* p3) {
	XAOL_AssertIfNULL_(p3);
	
	if (P3IsValid(p3)) {
		if (p3->data)
			MFree(p3->data);
		MFree(p3);
		p3 = NULL;
		}
	}
	
// Turn a P3 packet into a string
XAOLError P3BuildPacket (P3* p3, char** buffer, int* length) {
	DataStream* ds = NULL;
   
	XAOLError error;
	char* buf = NULL;
	int len;
	
    // SBW - 11/20 Changed for debugging
	ds = DSNew(NULL, 4096, true);
	if (ds == NULL)
		return (PE_DSError);
		
	error = DSPutUInt16(ds, p3->length + 3);
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}
		
	error = DSPutByte(ds, p3->seq);
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}
		
	error = DSPutByte(ds, p3->ack);
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}

	error = DSPutByte(ds, p3->type);
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}
		
	error = DSPutRaw(ds, (unsigned char*) p3->data, (int) p3->length);
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}
		
	// As of right here it has the packet ready
	error = DSGetBuffer(ds, &buf);
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}
		
	len = DSGetUsed(ds);
	
	p3->crc = CRC16(buf, len);
	
	DSDispose(ds);
	
        // SBW - Changed to true for debugging
	ds = DSNew(NULL, 4096, true);
	if (ds == NULL)
		return (PE_DSError);
	
	error = DSPutByte(ds, 'Z');
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}
		
	error = DSPutUInt16(ds, p3->crc);
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}
		
	error = DSPutRaw(ds, (unsigned char*) buf, len);
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}
		
	error = DSPutByte(ds, 13);
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}
		
	error = DSGetBuffer(ds, buffer);
	if (error != No_Error) {
		DSDispose(ds);
		return (error);
		}
		
	*length = DSGetUsed(ds);
	
	if (buf != NULL)
		MFree(buf);
	
	if (ds != NULL)
		DSDispose(ds);
	
	return (No_Error);
	}

// Parse a string out into a P3 Packet
XAOLError P3ParsePacket (P3* p3, char* buffer, int len, int* bytesLeft) {
	DataStream* ds;
	XAOLError dse;
	unsigned char magic;
	
	XAOL_AssertIfNULL_(p3);
	XAOL_AssertIfNULL_(*buffer);
	
	if (! P3IsValid(p3))
		return (PE_InvalidP3);
	
	if (buffer == NULL)
		return (PE_BadData);
	
	ds = DSNew(buffer, len, true);
	if (ds != NULL) {
		dse = DSGetByte(ds, &magic);
		XAOL_Assert_(dse != No_Error);
		if (dse != No_Error) {
			DSDispose(ds);
			return (dse);
			}
		
		// This is an ugly hack to handle nonstandard error messages in some P3 sessions
		if (magic != 'Z') {
			if (magic == 'X') {
				dse = DSGetByte(ds, &magic);
				XAOL_Assert_(dse != No_Error);
				if (magic == 'S') {
					DSDispose(ds);
					return (PE_PostMessage);
					}
				}
			else {
				DSDispose(ds);
				return (PE_BadData);
				}
			}
			
		dse = DSGetUInt16(ds, &p3->crc);
		XAOL_Assert_(dse != No_Error);
		if (dse != No_Error) {
			DSDispose(ds);
			return (dse);
			}
		
		dse = DSGetUInt16(ds, &p3->length);
		XAOL_Assert_(dse != No_Error);
		if (dse != No_Error) {
			DSDispose(ds);
			return (dse);
			}
		
		dse = DSGetByte(ds, &p3->seq);
		XAOL_Assert_(dse != No_Error);
		if (dse != No_Error) {
			DSDispose(ds);
			return (dse);
			}
		
		dse = DSGetByte(ds, &p3->ack);
		XAOL_Assert_(dse != No_Error);
		if (dse != No_Error) {
			DSDispose(ds);
			return (dse);
			}
		
		dse = DSGetByte(ds, &p3->type);
		if (dse != No_Error) {
			DSDispose(ds);
			return (dse);
			}
			
		dse = DSGetRaw(ds, (unsigned char**) &p3->data, p3->length - 2);
		XAOL_Assert_(dse != No_Error);
		if (dse != No_Error) {
			DSDispose(ds);
			return (dse);
			}
		
		*bytesLeft = len - DSGetUsed(ds);
			
		DSDispose(ds);
		}
	else
		return (Mem_Error);
	
	return (No_Error);
	}
